

#include <p18f4620.h>
#include <stdio.h>
#include <string.h>	// for memset
#include "defines.h"
#include "globals.h"
#include "gpib.h"
#include "db_utils.h"
#include "sio.h"
#include "SS80.h"
#include "pff.h"
//
// SS80 disk handling. See "Subset 80 from Fixed and flexible disc drives" 5958-4129 
//
// HP 9122 SS80 disk is stored on the SD card as a disk image cylinder - head - sector
//
// Docs:
// LIF filesystem http://www.hp9845.net/9845/projects/hpdir/#lif_filesystem

extern char string[MAX_DBG_STRING];

// ----------------------------------------- Structures used for testing ------------------------------------------
// HP/LIF data is BIG endian
//
// Volume label structure
//
#if defined(HP9122D)
ControllerDescriptionType ControllerDescription = {
0b10000000,
0b00000001,	// One unit
0x02,		// Transfer rate 744 = 0x2e8
0xe8,
0x05		// Controller type
};

UnitDescriptionType UnitDescription = {
0x01,		// Unit type = floppy
0x09,		// Type 9122
0x12,
0x20,
0x01,		// Bytes per block = 256
0x00,
0x01,		// Number of buffered blocks
0x00,		// Burst size = 0
0x17,		// Block time in us = 5888 = 0x1700
0x00,
0x00,		// Continous transfer rate = 45
0x2d,
0x11,		// Optimal retry = 4500 = 0x1194
0x94,
0x20,		// Access time = 8400 = 0x20d0
0xd0,
0x0f,		// Maximum interleave
0x00,		// Fixed volume byte (no fixed volumes)
0x01		// Removable volume byte = 1
};

VolumeDescriptionType VolumeDescription =
{
0x00,		// Max cylinder
0x00,
0x00,
0x00,		// Max head
0x00,		// Max sector
0x00,
0x00,		// Max single vector address = 2463 = 99f
0x00,
0x00,
0x00,
0x09,
0x9f,
0x02		// Current interleave = 2
};
#elif defined(HP9134L)
ControllerDescriptionType ControllerDescription = {
0b10000000,
0b00000001,	// One unit
0x02,		// Transfer rate 744 = 0x2e8
0xe8,
0x05		// Controller type
};

UnitDescriptionType UnitDescription = {
0x00,		// Unit type = winchester
0x09,		// Type 9134
0x13,
0x40,
0x01,		// Bytes per block = 256
0x00,
0x01,		// Number of buffered blocks
0x00,		// Burst size = 0
0x01,		// Block time in us = 502 = 0x1F6
0xF6,
0x00,		// Continous transfer rate = 140 kb/s
0x8c,
0x11,		// Optimal retry = 4500 = 0x1194 = 45000 = 4,5s worstcase
0x94,
0x11,		// Access time = 4500 = 0x1194
0x94,
0x1f,		// Maximum interleave
0x01,		// Fixed volume byte (1 fixed volumes)
0x00		// Removable volume byte = 0 (fixed)
};

VolumeDescriptionType VolumeDescription =
{
0x00,		// Max cylinder
0x00,
0x00,
0x00,		// Max head
0x00,		// Max sector
0x00,
0x00,		// Max single vector address = 58176 = e340
0x00,
0x00,
0x00,
0xe3,
0x40,
0x07		// Current interleave = 2
};
#else
#error You must define a floppy in defines.h
#endif
//
// Structure to hold head/cylinder/sector
//
DiskType Disk80 = {0,0,0,0};// Amigo disk addressing
BYTE Unit;
BYTE Volume;
BYTE Address[6];			// SS80 Address, Address[5] = LSB. In blocks
DWORD Vector;				// Vector, ie file position, in blocks
DWORD Maxvector;			// Maximum file position, used for bounds checking. Initialised by SS80_init.
LengthType Length;			// Length, in blocks
BYTE Errors;				// Error byte
//--------------------------------------------------------------------------------------------------------------------
extern unsigned char talking;
extern unsigned char listening;
unsigned volatile int SS80_tstate;		// Transaction state (idle/command/execution/report)
unsigned volatile int SS80_cstate;		// Command state state
unsigned char qstat;					// QSTAT byte, 0=OK, 1=Error, 2=Powerup
//
// Init SS80 state machine
//
void SS80_init()
{
	Unit = 0;
	Volume = 0;
	Vector = 0;
	Maxvector = VolumeDescription.V9 * 0x100000000L + 
				VolumeDescription.V10 * 0x1000000L + 
				VolumeDescription.V11 * 0x10000L + 
				VolumeDescription.V12 * 0x100L;
	sprintf(string,(const far rom char*)"\r\nMax Vector %ld (%lX) ", Maxvector, Maxvector);
	sio_puts(string);
	//
	talking = 0;
	listening = 0;
	SS80_tstate = SS80_TRAN_IDLE;
	SS80_cstate = SS80_START;
	qstat = 2;							// Indicates that the device has just been powered up.
	Errors = 0;
	EnablePP();
}
//
// This is the transaction and command parser. Parses byte by byte from gpib. Return a command token to stop GPIB reception and process a command
//
// Parsing is governed by two state-variables: cstate and tstate. tstate is the transaction state and is also used for non-transaction commands (transparent)
// cstate is the actual command state that handles reading of commands with parameters and also persists the command from command state to
// execution state
//
int SS80_char(unsigned char ch, unsigned char atn, unsigned char reset)
{
unsigned char status;

	if(reset) SS80_init();
	status = 0;
	//
	if(listening)
		sprintf(string,(const far rom char*)"\r\n{L%u %d}[%02X]", SS80_cstate,atn,ch);
	else if(talking)
		sprintf(string,(const far rom char*)"\r\n{T%u %d}[%02X]", SS80_cstate,atn,ch);
	else
		sprintf(string,(const far rom char*)"\r\n{%u %d}[%02X]", SS80_cstate, atn,ch);
	//sio_puts(string);
	//
	// Handle single-byte commands, they all reset the state-machine in various ways
	//
	if(atn)
	{
		if(ch == SDC || ch == DCL)						// Selected device clear or all device clear
		{
			status = SS80_CMD_CLEAR;
			SS80_tstate = SS80_TRAN_IDLE;
			SS80_cstate = SS80_START;
			return status;
		}
		if(ch == UNL)
		{
			listening = 0;
			SS80_tstate = SS80_TRAN_IDLE;
			return 0;
		}
		if(ch == UNT)
		{
			talking = 0;
			SS80_tstate = SS80_TRAN_IDLE;
			return 0;
		}
		if(ch >= 0x40 && ch <= 0x5e && ch != MTA)		// A talk address automatically untalks everyone
		{
			talking = 0;
			SS80_tstate = SS80_TRAN_IDLE;
			return 0;
		}
		if(ch == MTA)
		{
			talking = 1;
			SS80_tstate = SS80_TRAN_IDLE;
#if defined(SDEBUG)
			sprintf(string,(const far rom char*)"[MTA]");
			sio_puts(string);
#endif
			return 0;
		}
		if(ch == MLA)
		{
			listening = 1;
			SS80_tstate = SS80_TRAN_IDLE;
#if defined(SDEBUG)
			sprintf(string,(const far rom char*)"[MLA]");
			sio_puts(string);
#endif
			return 0;
		}
		// ---------------------------------------------------
		if(ch == 0x65 && listening)		// Go to command state
		{
			SS80_cstate = SS80_START;
			SS80_tstate = SS80_TRAN_COMMAND;
			return 0;
		}
		if(ch == 0x6E)		// Go to Execution state (both listen & talk)
		{
			SS80_tstate = SS80_TRAN_EXECUTION;
		}
		if(ch == 0x70 && talking)		// Go to report state
		{
			SS80_cstate = SS80_START;
			SS80_tstate = SS80_TRAN_REPORT;
		}
		if(ch == 0x70 && listening)		// This is Amigo clear
		{
			DisablePP();
			SS80_cstate = SS80_START;
			SS80_tstate = SS80_TRAN_AMIGO_CLEAR;
			return 0;					// Return and wait for next byte
		}
	}
	//
	//sprintf(string,(const far rom char*)"\r\nastate %d, ch %X", SS80_state, ch);
	//sio_puts(string);
	//
	switch(SS80_tstate)
	{
		case SS80_TRAN_IDLE:
			//
			break;
		case SS80_TRAN_COMMAND:
			status = SS80_do_command(ch);
			break;
		case SS80_TRAN_EXECUTION:
			DisablePP();
			status = SS80_exec_command(ch);
			break;
		case SS80_TRAN_REPORT:				// Report status
			DisablePP();					// Disable parallell poll
#if defined(SDEBUG)
			sprintf(string,(const far rom char*)"\r\nSS80:Rpt QSTAT %d ", qstat);
			sio_puts(string);
#endif
			memset(gpib_tx_buff,0,GPIB_MAX_TX_LEN);
			gpib_flags = gpib_flags | FLAGS_EOI;	// EOI on
			gpib_tx_buff[0] = qstat;
			sendgpib(1,0);
			gpib_flags = gpib_flags & ~FLAGS_EOI;	// EOI off
			status = SS80_CMD_NOP;
			break;

		case SS80_TRAN_AMIGO_CLEAR:			// Status byte from amigo clear. 0=Disable parity, 1=Enable parity
			EnablePP();
			status = 0;
#if defined(SDEBUG)
			sprintf(string,(const far rom char*)"\r\nSS80:Amigo Clear %d ", ch);
			sio_puts(string);
#endif
			break;
		default:
			break;
	}
	//sprintf(string,(const far rom char*)"(%d)", status);
	//sio_puts(string);
	return status;
}
//
// Act on a SS/80 command
//
BYTE SS80_do_command(BYTE ch)
{
BYTE status = 0;

	switch(SS80_cstate)
	{
		case SS80_START:
			if(ch == 0x00 && listening)		// Read
			{
				sprintf(string,(const far rom char*)"\r\nSS80:Read");
				sio_puts(string);
				status = 0;
				SS80_cstate = SS80_000;
				EnablePP();					// Enable PP triggers the host to go to execution mode
				break;
			}
			if(ch == 0x02 && listening)		// Write
			{
				sprintf(string,(const far rom char*)"\r\nSS80:Write");
				sio_puts(string);
				status = 0;
				SS80_cstate = SS80_002;
				EnablePP();					// Enable PP triggers the host to go to execution mode
				break;
			}
			if(ch >= 0x20 && ch <= 0x2f && listening)		// Set unit
			{
				Unit = ch - 0x20;
				sprintf(string,(const far rom char*)"\r\nSS80:Set unit %d", Unit);
				sio_puts(string);
				status = 0;
				SS80_cstate = SS80_START;
				//EnablePP();					// Enable PP triggers the host to go to execution mode
				break;
			}
			if(ch >= 0x40 && ch <= 0x4f && listening)		// Set unit
			{
				Volume = ch - 0x40;
				sprintf(string,(const far rom char*)"\r\nSS80:Set volume %d", Volume);
				sio_puts(string);
				status = 0;
				SS80_cstate = SS80_START;
				//EnablePP();					// Enable PP triggers the host to go to execution mode
				break;
			}
//			if(ch == 0x34 && listening)		// No operation
//			{
//				SS80_cstate = SS80_034;
//				status = 0;
//				break;
//			}
			if(ch == 0x35 && listening)		// Describe
			{
				sprintf(string,(const far rom char*)"\r\nSS80:Describe");
				sio_puts(string);
				status = 0;
				SS80_cstate = SS80_035;
				EnablePP();					// Enable PP triggers the host to go to execution mode
				break;
			}
//			if(ch == 0x37 && listening)		// Initialise media, 2 bytes follow
//			{
//				SS80_cstate = SS80_037_1;
//				break;
//			}
			if(ch == 0x0d && listening)		// Request status
			{
				status = SS80_CMD_SEND_STATUS;
				SS80_cstate = SS80_00D;
				sprintf(string,(const far rom char*)"\r\nSS80:Rq status ");
				sio_puts(string);
				break;
			}
			if(ch == 0x10 && listening)		// Set address, 6 bytes follow
			{
				SS80_cstate = SS80_010_1;
				break;
			}
			if(ch == 0x18 && listening)		// Set length, 4 bytes follow
			{
				SS80_cstate = SS80_018_1;
				break;
			}
			sprintf(string,(const far rom char*)"\r\nSS80:Unknown %X ",ch);
			sio_puts(string);
			break;

		case SS80_010_1:
			Address[0] = ch;	// MSB
			SS80_cstate = SS80_010_2;
			break;
		case SS80_010_2:
			Address[1] = ch;
			SS80_cstate = SS80_010_3;
			break;
		case SS80_010_3:
			Address[2] = ch;
			SS80_cstate = SS80_010_4;
			break;
		case SS80_010_4:
			Address[3] = ch;
			SS80_cstate = SS80_010_5;
			break;
		case SS80_010_5:
			Address[4] = ch;
			SS80_cstate = SS80_010_6;
			break;
		case SS80_010_6:					// TODO. Only handles 4-byte vectors, 0xffffffff bytes = 4 294 967 295 bytes
			Address[5] = ch;
			Vector = Address[2] * 0x100000000L + Address[3] * 0x1000000L + Address[4] * 0x10000L + Address[5] * 0x100L;
			sprintf(string,(const far rom char*)"\r\nSS80:Set address %lX ",Vector);
			sio_puts(string);
			SS80_cstate = SS80_START;	// and back to the top for next cmd
			break;

		case SS80_018_1:
			Length.Lenbytes[3] = ch;	// MSB
			SS80_cstate = SS80_018_2;
			break;
		case SS80_018_2:
			Length.Lenbytes[2] = ch;
			SS80_cstate = SS80_018_3;
			break;
		case SS80_018_3:
			Length.Lenbytes[1] = ch;
			SS80_cstate = SS80_018_4;
			break;
		case SS80_018_4:
			Length.Lenbytes[0] = ch;	// LSB
			sprintf(string,(const far rom char*)"\r\nSS80:Set length %lX ",Length.Length);
			sio_puts(string);
			SS80_cstate = SS80_START;
			break;

		case SS80_034:					// No operation	
			sprintf(string,(const far rom char*)"\r\nSS80:NOP");
			sio_puts(string);
			SS80_cstate = SS80_START;		
			break;

		case SS80_035:					
			break;
										// Initialise disk command
		case SS80_037_1:				// First byte of init params
			SS80_cstate = SS80_037_2;
			break;
		case SS80_037_2:				// Second byte of init param
			sprintf(string,(const far rom char*)"\r\nSS80:Init disk");
			sio_puts(string);
			SS80_cstate = SS80_START;
			status = 0;
			EnablePP();
			break;
		case SS80_END:
			break;
	}
	return status;
}
//
// Execute a SS/80 command
//
BYTE SS80_exec_command(BYTE ch)
{
BYTE status;

	switch(SS80_cstate)
	{
		case SS80_000:
			status = SS80_CMD_BUFFERED_READ;
			break;
		case SS80_002:
			status = SS80_CMD_BUFFERED_WRITE;
			SS80_cstate = SS80_START;
			break;
		case SS80_035:
			status = SS80_CMD_DESCRIBE;
			break;
		case SS80_00D:
			status = SS80_CMD_SEND_STATUS;
			break;
		case SS80_END:
			break;
		default:
			break;
	}
	return status;
}
//
// Return drive parameters to host
//
void SS80_cmd_describe()
{
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nSS80:Describe ");
	sio_puts(string);
#endif
	//
	gpib_flags = gpib_flags & ~FLAGS_EOI;	// No EOI
	memcpy(gpib_tx_buff,(void *)&ControllerDescription,5);
	sendgpib(5,0);
	//
	memcpy(gpib_tx_buff,(void*)&UnitDescription,19);
	sendgpib(19,0);
	//
	gpib_flags = gpib_flags | FLAGS_EOI;	// EOI on
	memcpy(gpib_tx_buff,(void *)&VolumeDescription,13);
	sendgpib(13,0);
	//
	gpib_flags = gpib_flags & ~FLAGS_EOI;
	EnablePP();
}
//
// Return Status to host
//
// Only status we bother returning for now is write protect
//
void SS80_cmd_send_status()
{
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nSS80:Status ");
	sio_puts(string);
#endif
	//
	gpib_flags = gpib_flags & ~FLAGS_EOI;	// No EOI
	memset(gpib_tx_buff,0,GPIB_MAX_TX_LEN);
	gpib_tx_buff[0] = 0x11;				// Volume 1, Unit 1
	gpib_tx_buff[1] = 0xff;
	//
	if(Errors & SS80_ERR_SEEK) if(WP_PORT) gpib_tx_buff[4] = 0b00000001; // address bounds = Byte 3
	if(Errors & SS80_ERR_READ) if(WP_PORT) gpib_tx_buff[6] = 0b00000010; // unit fault = Byte 5
	if(Errors & SS80_ERR_WRITE) if(WP_PORT) gpib_tx_buff[6] = 0b00000010; // unit fault = Byte 5
	if(WP_PORT) gpib_tx_buff[6] = 0b00001000; // Write protect = Byte 7
	//
	gpib_tx_buff[11] = Address[0];		// Byte 11-16 = target address
	gpib_tx_buff[12] = Address[1];
	gpib_tx_buff[13] = Address[2];
	gpib_tx_buff[14] = Address[3];
	gpib_tx_buff[15] = Address[4];
	gpib_tx_buff[16] = Address[5];
	sendgpib(20,0);

	gpib_flags = gpib_flags & ~FLAGS_EOI;
	EnablePP();
}
//
// Seek to the position in vector (called by read and write). Returns nonzero on error
//
int SS80_cmd_seek()
{
DWORD fpos;
FRESULT rc;
	fpos = Vector;
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nSS80: Seek: ");
	sio_puts(string);
	sprintf(string,(const far rom char*)"\r\nSeeking to %ld (%lX) ", fpos, fpos);
	sio_puts(string);
#endif
	//
	// Bounds checking
	//
	if (Vector > Maxvector)
	{
		sprintf(string,(const far rom char*)"\r\nSeek beyond end");
		sio_puts(string);
		qstat = 1;
		Errors = Errors | SS80_ERR_SEEK;		
	}
	if ((Vector + Length.Length) > Maxvector)
	{
		sprintf(string,(const far rom char*)"\r\nRead beyond end");
		sio_puts(string);
		qstat = 1;
		Errors = Errors | SS80_ERR_SEEK;		
	}
	qstat = 0;
	rc=pf_lseek(fpos);
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nSeek to %ld failed %d", fpos, rc);
		sio_puts(string);
		qstat = 1;
		Errors = Errors | SS80_ERR_SEEK;
	}
	return qstat;
}
//
// Do a buffered read of the current sector
//
// Data is (for now) always sent in 256-bit chunks, because of how the GPIB TX buffer was once set up. So for now
// we calculate how many 256-bit chunks to send and then send the rest. Return nonzero on error
//
int SS80_cmd_buffered_read()
{
int i,rc2;
DWORD total_bytes;
FRESULT rc;
WORD nbytes;
int chunks;				// Number of 64 byte chunks to send
int tail;				// tail after the chunks
						// Send Length bytes. EOI on last packet

	DisablePP();
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nSS80:Read %ld (%lX) ", Vector, Vector);
	sio_puts(string);
	//Time();
#endif
	rc2 = SS80_cmd_seek();
	if(rc2 != 0)
	{
		Errors = Errors | SS80_ERR_SEEK;
		SS80_error_return();		// return one byte and
		return rc2;					// bail out if seek fails
	}
	//
	//								// Calculate how to send our data
	chunks = Length.Length / 256;	// Whole 256-byte blocks
	tail = Length.Length % 256;		// and possible tail < 256
	//
	gpib_flags = gpib_flags & ~FLAGS_EOI;
	total_bytes = 0;
	//Time();
	for(i=0;i<chunks;i++)	// Read maximum 256 bytes
	{
		if(i==(chunks-1) && tail == 0) gpib_flags = gpib_flags | FLAGS_EOI;	// EOI on, on last block if there is no tail
		rc=pf_read(gpib_tx_buff,256,&nbytes);
#if defined(SDEBUG)
		sprintf(string,(const far rom char*)"\r\n1:Read rtn %X, %d read ", rc, nbytes);
		sio_puts(string);
#endif
		if(rc != 0)
		{
			Errors = Errors | SS80_ERR_SEEK;
			SS80_error_return();		// return one byte and
			return rc2;					// bail out if seek fails
		}
		//Time();
		sendgpib(nbytes,0);
		//Time();
		total_bytes = total_bytes + nbytes;
		//DumpData(gpib_tx_buff, nbytes);
		Vector = Vector + 0x100L;
		sprintf(string,(const far rom char*)"\r\nVector %ld (%lX) ", Vector, Vector);
		sio_puts(string);
	}
	// Send the rest, in case it did not fit in the chunks
	if(tail > 0)
	{
		rc=pf_read(gpib_tx_buff,tail,&nbytes);
#if defined(SDEBUG)
		sprintf(string,(const far rom char*)"\r\n2:Read rtn %X, %d read ", rc, nbytes);
		sio_puts(string);
#endif
		if(rc != 0)
		{
			Errors = Errors | SS80_ERR_SEEK;
			SS80_error_return();		// return one byte and
			return rc2;					// bail out if read fails
		}
		gpib_flags = gpib_flags | FLAGS_EOI;	// EOI on, on tail block
		//Time();
		sendgpib(tail,0);
		//Time();
		total_bytes = total_bytes + nbytes;
		//DumpData(gpib_tx_buff, tail);
		Vector = Vector + 0x100L;
		sprintf(string,(const far rom char*)"\r\nVector %ld (%lX) ", Vector, Vector);
		sio_puts(string);
	}
	gpib_flags = gpib_flags & ~FLAGS_EOI;
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\n %ld bytes sent ", total_bytes);
	sio_puts(string);
//Time();
#endif
	qstat = 0;
	EnablePP();
	return qstat;
}
//
// Receive data and call write_sector() for each chunk of 256 bytes. returns nonzero on error
//
int SS80_cmd_buffered_write()
{
int more,i,j;
DWORD total_bytes;
FRESULT rc;
int rc2;		// readgpib() returns an integer, silly!
int nbytes;

	qstat = 0;
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nSS80:Write %ld (%lX) ", Vector, Vector);
	sio_puts(string);
#endif
	rc2 = SS80_cmd_seek();
	if(rc2 != 0) return rc2;		// Bail out if seek fails
	//
	total_bytes = 0;
	more = 1;

	while(more)							// Loop until we have received Length sectors or EOI
	{
		gpib_ptr = 0;
		rc2=readgpib();
#if defined(SDEBUG)
		if(gpib_ptr > 0)
		{
			sprintf(string,(const far rom char*)"\r\nGPIB rc:%d %d bytes\r\n", rc2,gpib_ptr);
			sio_puts(string);
		}
#endif
		//	-----------------------------------
		total_bytes = total_bytes + gpib_ptr;
		if(total_bytes >= Length.Length) more = 0;	// All done?
		if(rc2==254) more = 0; 						// EOI		
		if(IFC_PORT == 0) more = 0;					// IFC
		//
		// SS/80 behaviour: Replicate last byte in case it is a partial write
		//
		if(gpib_ptr < 256)
		{
			memset(&gpib_buff[gpib_ptr],gpib_buff[gpib_ptr-1],256-gpib_ptr);
		}
		rc=Write_sector();
		if(rc)
		{
			sprintf(string,(const far rom char*)"\r\nWrite_sectors fail %d", rc);
			sio_puts(string);
			qstat = 1;
			Errors = Errors | SS80_ERR_SEEK;
			return qstat;
		}
		//sprintf(string,(const far rom char*)"\r\nWr returns %X, %d bytes wr", rc, nbytes);
		//sio_puts(string);
	}
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nSS80:Buff wr %ld bytes rx", total_bytes);
	sio_puts(string);
#endif
							// The address is incremented on a successful write by Write_sectors().
	if(WP_PORT) qstat = 1;
	EnablePP();
	return 0;
}
//
// QSTAT, return status byte
//
void SS80_cmd_qstat()
{
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nAM:DSJ");
	sio_puts(string);
#endif
	memset(gpib_tx_buff,0,GPIB_MAX_TX_LEN);
	gpib_tx_buff[0] = qstat;
	sendgpib(1,0);	
	//
	// QSTAT clears the qstat byte that might be 2 after powerup
	//
	qstat = 0;
}
//
// Response to selected device clear or all clear - Select sector 0
//
//
void SS80_cmd_clear()
{
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nSS80:Clear");
	sio_puts(string);
#endif
	Vector = 0;
	Errors = 0;
	Length.Length = 0;
	EnablePP();
	//
	// Clear the DSJ byte that might be 2 after powerup
	//
	qstat = 0;
}
//
// Increment vector
//
void SS80_increment()
{
DWORD fpos;
FRESULT rc;

	Vector = Vector + 0x100L;
	fpos = Vector;
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nSS80: Increment:");
	sio_puts(string);
	sprintf(string,(const far rom char*)"\r\nSeeking to %ld (%lX)", fpos, fpos);
	sio_puts(string);
#endif
	rc=pf_lseek(fpos);
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nSeek to %ld failed %d", fpos, rc);
		sio_puts(string);
		qstat = 1;
	}
	else qstat = 0;
}

int Write_sector()
{
// Write 256 bytes in gpib_buff at current position, then advance to next sector
//
// Note that PetitFATFs only handles writes of full 512-byte sectors, so additional logic is needed.
// Sectors are nubered sequentially, from 0 to 66*2*16. Even numbered sectors are at the start of each 512-byte
// disk sector and odd numbered sectors at the end. Thus the procedure to write an even sector becomes:
// Read the upper 256 bytes into a buffer, seek to start of 512-byte sector, write 256 byte SS80 sector, then write buffer.
//
// For odd numbered sectors, we seek to the start of a 512-byte sector, read 256 bytes into a buffer
// seek back to start of sector, write the 256 bytes we just read, then write the SS80 (256byte) sector.
//
// Returns 0 on success, nonzero otherwise
//
int i,j,rc2;
DWORD fpos, sector;
WORD nbytes;
FRESULT rc;
int status;

	sector = Vector / 0x100L;
	status = 0;
#if defined(SDEBUG)
	sprintf(string,(const far rom char*)"\r\nWrite %ld (%lX) s:%ld ", Vector, Vector,sector);
	sio_puts(string);
	DumpData(gpib_buff, 256);
#endif
	rc2 = SS80_cmd_seek();
	if(rc2 != 0) return rc2;		// Bail out if seek fails
	//
	if(sector % 2 == 0)	// even, Read the upper part of the 512-byte sector
	{
		fpos = (sector / 2) * 512L + 256L;
		rc=pf_lseek(fpos);
		if(rc != 0) return rc;		// Bail out if seek fails
		rc=pf_read(buffer,256,&nbytes);
		if(rc != 0) return rc;		// Bail out if read fails
#if defined(SDEBUG)
		sprintf(string,(const far rom char*)"\r\nRd odd sector: %ld at %lX %d\r\n", sector+1, fpos, rc);
		sio_puts(string);
#endif
		fpos = fpos - 256L;
		rc=pf_lseek(fpos);	// Seek back
		if(rc != 0) return rc;		// Bail out if seek fails
	}
	else
	{
		// Read the lower part of the 512-byte sector,
		fpos = sector / 2 * 512L;
		rc=pf_lseek(fpos);
		if(rc != 0) return rc;		// Bail out if seek fails
		rc=pf_read(buffer,256,&nbytes);

#if defined(SDEBUG)
		sprintf(string,(const far rom char*)"\r\nRd even sector: %ld at %lX %d\r\n", sector-1, fpos, rc);
		sio_puts(string);
#endif
		if(rc != 0) return rc;		// Bail out if read fails
		// Seek back
		rc=pf_lseek(fpos);
		if(rc != 0) return rc;		// Bail out if seek fails
		// Write lower part
		if(!WP_PORT) rc=pf_write((void*)buffer, 256, &nbytes);
		if(rc != 0) return rc;		// Bail out if write fails
	}

	if(!WP_PORT) rc=pf_write((void*)gpib_buff, 256, &nbytes);	// Always write 256 bytes, ie a full block
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nWrite fail %d", rc);
		sio_puts(string);
		status = 1;
		return rc;
	}
		//sprintf(string,(const far rom char*)"\r\nWr returns %X, %d bytes wr", rc, nbytes);
		//sio_puts(string);

	if(sector % 2 == 0)	// even
	{
		// Write the upper part of the 512-byte sector
		if(!WP_PORT) rc=pf_write((void*)buffer, 256, &nbytes);
		if(rc != 0) return rc;		// Bail out if write fails
	}
	if(!WP_PORT) rc = pf_write(0, 0, &nbytes);
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nClosing wr fail %d", rc);
		sio_puts(string);
		status = 1;
		return rc;
	}
	else
	{
		//sprintf(string,(const far rom char*)"\r\nClosed sector");
		//sio_puts(string);
	}
	if(status == 0) SS80_increment();
	return status;
}
//
// Error return. On catastrophic failures, just return one byte, Value 1, with EOI
//
void SS80_error_return()
{
	memset(gpib_tx_buff,0,GPIB_MAX_TX_LEN);
	gpib_tx_buff[0] = 1;
	gpib_flags = gpib_flags | FLAGS_EOI;	// EOI on
	sendgpib(1,0);	
	gpib_flags = gpib_flags & ~FLAGS_EOI;
}
//------------------------- Test routines --------------------
//char TD[] = {0x3f,1,0x41,1,0x20,1,0x65,1,
//0x35,0,
//0x3f,1,
//0x3f,1,
//0x21,1,
//0x40,1,
//0x6e,1, // exec phase secondary
//0x5f,1,
//0x3f,1,
//0x21,1,
//0x40,1,
//0x70,1,	// Rpt phase secondary
//0x3F,1, // UNL
//0x5F,1	// UNT
//};

char TD[] = {
0x3F,0,
0x5f,1,
0x20,1,
0x65,1,
0x20,0,
0x40,0, 
0x34,0, // NOP
0x37,0, // Format
0x00,0,
0x0a,0
};
void SS80_Test()
{
int i;
BYTE rc;
Length.Lenbytes[1] = 2;
Address[5] = 2;
Vector = Address[2] * 0x100000000L + Address[3] * 0x1000000L + Address[4] * 0x10000L + Address[5] * 0x100L;
for(i=0;i<sizeof(TD);i=i+2)
	{
		rc=SS80_char(TD[i],TD[i+1],0);
		sprintf(string,(const far rom char*)"\r\nrc=%d",rc);
		sio_puts(string);
	}
}