#include <p18f4620.h>
#include <stdio.h>
#include <string.h>	// for memset
#include "globals.h"
#include "gpib.h"
#include "db_utils.h"
#include "sio.h"
#include "amigo.h"
#include "pff.h"
//
// Amigo disk handling. See "Subset 80 from Fixed and flexible disc drives" 5958-4129 and the 9123 disc drive manual addendum 5957-6584
//
// HP 9121 Amigo disk is stored on the SD card as a disk image cylinder - head - sector
//
// Docs:
// LIF filesystem http://www.hp9845.net/9845/projects/hpdir/#lif_filesystem

extern char string[MAX_DBG_STRING];

// ----------------------------------------- Structures used for testing ------------------------------------------
// LIF data is BIG endian
//
// Volume label structure
//
VolumeLabelType vl = {0x0080,
'L','A','B','E','L','1',
0,0x200,
0x0080,
0,
0,0x0e00,						// Change this to 0x0e -> LIF directory too big
0
};
//
// Directory entry
//
DirEntryType de = {'M','Y','F','I','L','E','1','2','3','4',0x100,0,0x100,0,0x20,0x13,0x10,0x12,0x00,0x00,0x00,1,0,0};
//
// Structure to hold head/cylinder/sector
//
DiskType Disk = {0,0,0,0};
//--------------------------------------------------------------------------------------------------------------------
extern near unsigned char talking;
extern near unsigned char listening;
unsigned volatile int amigo_state;
unsigned char dsj;
//
// Init Amigo state machine
//
void amigo_init()
{
	talking = 0;
	listening = 0;
	amigo_state = AMIGO_START;
	dsj = 2;
	EnablePP();
}
//
// This is the command parser. Parses byte by byte from gpib
//
int amigo_char(unsigned char ch, unsigned char atn, unsigned char reset)
{
unsigned char status;

	if(reset) amigo_init();
	status = 0;
	//
	// Strip parity bit
	//
	ch = ch & 0x7F;
	//
#if SDEBUG > 2
	if(listening)
		sprintf(string,(const far rom char*)"{L%u %d}[%02X]", amigo_state,atn,ch);
	else if(talking)
		sprintf(string,(const far rom char*)"{T%u %d}[%02X]", amigo_state,atn,ch);
	else
		sprintf(string,(const far rom char*)"{%u %d}[%02X]", amigo_state, atn,ch);
	sio_puts((unsigned char*)string);
#endif
	//
	// Handle single-byte commands, they all reset the state-machine
	//
	if(atn)
	{
		if(ch == SDC || ch == DCL)						// Selected device clear or all device clear
		{
			status = AMIGO_CMD_CLEAR;
			amigo_state = AMIGO_START;
			return status;
		}
		if(ch == UNL)
		{
			listening = 0;
			amigo_state = AMIGO_START;
			return 0;
		}
		if(ch == UNT)
		{
			talking = 0;
			amigo_state = AMIGO_START;
			return 0;
		}
		if(ch >= 0x40 && ch <= 0x54 && ch != MTA)		// A talk address automatically untalks everyone
		{
			talking = 0;
			amigo_state = AMIGO_START;
			return 0;
		}
		if(ch == MTA)
		{
			talking = 1;
			amigo_state = AMIGO_START;
#if SDEBUG > 1
			sprintf(string,(const far rom char*)"[MTA]");
			sio_puts(string);
#endif
			return 0;
		}
		if(ch == MLA)
		{
			listening = 1;
			amigo_state = AMIGO_START;
#if SDEBUG > 1
			sprintf(string,(const far rom char*)"[MLA]");
			sio_puts(string);
#endif
			return 0;
		}
	}
	//
	//sprintf(string,(const far rom char*)"\r\nastate %d, ch %X", amigo_state, ch);
	//sio_puts((unsigned char*)string);
	//
	switch(amigo_state)
	{
		case AMIGO_START:
			//
			if(ch == 0x60 && talking)		// Single byte command
			{
				//sprintf(string,(const far rom char*)"\r\nAM:Buffered read", Disk.sector);
				//sio_puts((unsigned char*)string);
				status = AMIGO_CMD_BUFFERED_READ;
				amigo_state = AMIGO_START;
				break;
			}
			if(ch == 0x60 && listening)		// Single byte command
			{
				//sprintf(string,(const far rom char*)"\r\nBuffered write", Disk.sector);
				//sio_puts((unsigned char*)string);
				status = AMIGO_CMD_BUFFERED_WRITE;
				amigo_state = AMIGO_START;
				break;
			}
			if(ch == 0x68 && listening) amigo_state = AMIGO_068_L;
			if(ch == 0x68 && talking)		// Single byte cmd
			{
				//sprintf(string,(const far rom char*)"\r\nSend status", Disk.sector);
				//sio_puts((unsigned char*)string);
				status = AMIGO_CMD_SEND_STATUS;
				amigo_state = AMIGO_START;
				break;
			}
			if(ch == 0x6A && listening) amigo_state = AMIGO_06A;
			if(ch == 0x70 && talking) 	// DSJ is a single byte command
			{
				//sprintf(string,(const far rom char*)"\r\nDSJ");
				//sio_puts((unsigned char*)string);
				status = AMIGO_CMD_DSJ;
				amigo_state = AMIGO_START;
				DisablePP();
				break;
			}
			break;

		case AMIGO_068_L:
			if(ch == 0x02) amigo_state = AMIGO_068_SEEK;
			if((ch == 0x03) || (ch == 0xff))
			{
				DisablePP();
				status = 0; // Do nothing just a request wait for talk. Enable PP and wait for talk, then the 0x68 cmd
				EnablePP();
				amigo_state = AMIGO_START;
				break;
			}
			if(ch == 0x16) amigo_state = AMIGO_068_WAKEUP;
			break;
		case AMIGO_068_T:
			//sprintf(string,(const far rom char*)"\r\nSend status", Disk.sector);
			//sio_puts((unsigned char*)string);
			status = AMIGO_CMD_SEND_STATUS;
			amigo_state = AMIGO_START;
			break;
		case AMIGO_06A:
			if(ch == 0x05) amigo_state = AMIGO_06A_READ;
			break;
		case AMIGO_068_SEEK:					// seek, five bytes follow unknown, unit, head,sector
			amigo_state = AMIGO_068_SEEK_1;
			break;
		case AMIGO_068_SEEK_1:
			Disk.unit = ch;
			amigo_state = AMIGO_068_SEEK_2;
			break;
		case AMIGO_068_SEEK_2:
			Disk.cyl = ch;
			amigo_state = AMIGO_068_SEEK_3;
			break;
		case AMIGO_068_SEEK_3:
			Disk.head = ch;
			amigo_state = AMIGO_068_SEEK_4;
			break;
		case AMIGO_068_SEEK_4:
			Disk.sector = ch;
			//sprintf(string,(const far rom char*)"\r\nSeek: %d %d %d %d, %X", Disk.unit, Disk.cyl, Disk.head, Disk.sector, ch);
			//sio_puts((unsigned char*)string);
			amigo_state = AMIGO_START;
			status = AMIGO_CMD_SEEK;
			break;
			//
		case AMIGO_068_WAKEUP:
			amigo_state = AMIGO_START;
			status = AMIGO_CMD_WAKEUP;
			break;							
		case AMIGO_06A_READ:							// Does nothing, but prepares for the read. Actual read bytes is AMIGO_60
			amigo_state = AMIGO_06A_READ_2;				//two bytes follow <- Wrong, one byte, skip to read2
			break;
		case AMIGO_06A_READ_1:							// Read byte 1
			amigo_state = AMIGO_06A_READ_2;
			break;
		case AMIGO_06A_READ_2:							// Read byte 2
			EnablePP();
			amigo_state = AMIGO_END;
			break;
		case AMIGO_070_T:									// This is DSJ, put status byte on bus
			//sprintf(string,(const far rom char*)"\r\nDSJ");
			//sio_puts((unsigned char*)string);
			amigo_state = AMIGO_START;
			status = AMIGO_CMD_DSJ;
			amigo_state = AMIGO_END;
			break;
		case AMIGO_070_L:									// This is Amigo clear. Seek to 0/0/0 and put status byte on bus
		case AMIGO_END:
		default:
			break;
	}
	//sprintf(string,(const far rom char*)"(%d)", status);
	//sio_puts((unsigned char*)string);
	return status;
}
//
// Return drive status to host. TODO: Fix stuff like proper status for missing SD etc
//
void amigo_cmd_send_status()
{
#if SDEBUG > 0
	sprintf(string,(const far rom char*)"\r\nAM:Send status");
	sio_puts(string);
#endif
	//
	gpib_flags = gpib_flags | FLAGS_EOI;
	gpib_tx_buff[0] = 0x00;				// Status 1
	if(WP_PORT) gpib_tx_buff[2] = gpib_tx_buff[2] | 0b00010011;	// Status 2 error
	gpib_tx_buff[1] = 0x00;				//
	gpib_tx_buff[2] = 0x0d;				// Status 2 0010 = hp format, 1=HP9121
	gpib_tx_buff[3] = 0x00;				//
	if(WP_PORT) gpib_tx_buff[3] = 0x60;	// Write protect
	sendgpib(4);
	gpib_flags = gpib_flags & ~FLAGS_EOI;
	EnablePP();
}
//
// Seek to the position held in the disk structure
//
void amigo_cmd_seek()
{
DWORD fpos;
FRESULT rc;
	fpos = (DWORD)Disk.cyl*16L*256L*2L + (DWORD)Disk.head * 256L *16L + (DWORD)Disk.sector *256L;
#if SDEBUG > 0
	sprintf(string,(const far rom char*)"\r\nAM: Seek: %d %d %d %d, s:%ld", 
		Disk.unit, Disk.cyl, Disk.head, Disk.sector,
		(DWORD)Disk.cyl*16L*2L + (DWORD)Disk.head  *16L + (DWORD)Disk.sector
		);
	sio_puts(string);
	sprintf(string,(const far rom char*)"\r\nSeeking to %ld (%lX)", fpos, fpos);
	sio_puts(string);
#endif
	rc=pf_lseek(fpos);
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nSeek to %ld failed %d", fpos, rc);
		sio_puts(string);
	}
	dsj = 0;
	EnablePP();
}
//
// Do a buffered read of the current sector
//
void amigo_cmd_buffered_read()
{
int i;
DWORD total_bytes;
FRESULT rc;
WORD nbytes;
						// Follows after 0x680x05, Send  256 bytes. EOI only if error
#if SDEBUG > 0
	sprintf(string,(const far rom char*)"\r\nAM:Buff rd c/h/s: %d/%d/%d s:%ld ", Disk.cyl, Disk.head, Disk.sector,
		(DWORD)Disk.cyl*16L*2L + (DWORD)Disk.head  *16L + (DWORD)Disk.sector
	);
	sio_puts(string);
#endif
	//
	total_bytes = 0;
	//for(i=0;i<4;i++)	// Read 256 bytes
	//{
		rc=pf_read(gpib_tx_buff,256,&nbytes);
#if SDEBUG > 0
		sprintf(string,(const far rom char*)"\r\nRead returns %X, %d bytes read", rc, nbytes);
		sio_puts(string);
#endif

		gpib_flags = gpib_flags & ~FLAGS_EOI;
		//gpib_tx_buff[0] = ' ';
		//gpib_tx_buff[1] = nbytes+2;
		sendgpib(nbytes);
		total_bytes = total_bytes + nbytes;
	//}
#if SDEBUG > 0
	sprintf(string,(const far rom char*)"\r\n %ld bytes sent", total_bytes);
	sio_puts(string);
#endif
	dsj = 0;
	EnablePP();	
}
//
// Write 256 bytes at current position, then advance to next sector
//
// Note that PetitFATFs only handles writes of full 512-byte sectors, so additional logic is needed.
// Sectors are nubered sequentially, from 0 to 66*2*16. Even numbered sectors are at the start of each 512-byte
// disk sector and odd numbered sectors at the end. Thus the procedure to write an even sector becomes:
// Read the upper 256 bytes into a buffer, seek to start of 512-byte sector, write 256 byte Amigo sector, then write buffer.
//
// For odd numbered sectors, we seek to the start of a 512-byte sector, read 256 bytes into a buffer
// seek back to start of sector, write the 256 bytes we just read, then write the Amigo (256byte) sector.
//
void amigo_cmd_buffered_write()
{
int more,i,j;
DWORD total_bytes;
DWORD fpos, sector;
FRESULT rc;
WORD nbytes;

						// Follows after 0x690x08, receive  max 256 bytes or up to EOI
#if SDEBUG > 0
	sprintf(string,(const far rom char*)"\r\nAM:Buff wr c/h/s: %d/%d/%d s:%ld", Disk.cyl, Disk.head, Disk.sector,
		(DWORD)Disk.cyl*16L*2L + (DWORD)Disk.head  *16L + (DWORD)Disk.sector
	);
	sio_puts(string);
#endif
	//
	sector = (DWORD)Disk.cyl*16L*2L + (DWORD)Disk.head  *16L + (DWORD)Disk.sector;
	total_bytes = 0;
	more = 1;
	if(sector % 2 == 0)	// even
	{
		// Read the upper part of the 512-byte sector
		fpos = (sector / 2) * 512L + 256L;
		rc=pf_lseek(fpos);
		rc=pf_read(buffer,256,&nbytes);
#if SDEBUG > 0
		sprintf(string,(const far rom char*)"\r\nRd odd sector: %ld at %lX %d", sector+1, fpos, rc);
		sio_puts(string);
#endif
		fpos = fpos - 256L;
		rc=pf_lseek(fpos);	// Seek back
	}
	else
	{
		// Read the lower part of the 512-byte sector,
		fpos = sector / 2 * 512L;
		rc=pf_lseek(fpos);
		rc=pf_read(buffer,256,&nbytes);
#if SDEBUG > 0
		sprintf(string,(const far rom char*)"\r\nRd even sector: %ld at %lX %d", sector-1, fpos, rc);
		sio_puts(string);
#endif
		// Seek back
		rc=pf_lseek(fpos);
		// Write lower part
		if(!WP_PORT) rc=pf_write((void*)buffer, 256, &nbytes);
	}
	while(more)
	{
		gpib_ptr = 0;
		rc=readgpib();
		//	TODO: Check for cpib_ptr>0
		//if(gpib_ptr > 0) DumpData(gpib_buff,gpib_ptr);
//		for(i=0;i<gpib_ptr;i++)
//		{
//			sprintf(string,(const far rom char*)"[%2X]", gpib_buff[i]);
//			sio_puts(string);		
//		}
		//--------------------------------------
		// Dump debug data
		//
#if SDEBUG > 1
		for(j=0;j<80&&(j*16<=gpib_ptr);j++){
			sprintf(string,(const far rom char*)"\r\n");
			sio_puts(string);
			for(i=0;i<16 && (i+j*16<=gpib_ptr);i++)
				{
					sprintf(string, (const far rom char*)" %02X",gpib_buff[i+j*16]);
					sio_puts(string);
				}
			sprintf(string, (const far rom char*)" | ");
			sio_puts(string);
			for(i=0;i<16 && (i+j*16<=gpib_ptr);i++)
				{
					if(gpib_buff[i+j*16]>' ') sprintf(string, (const far rom char*)"%c",gpib_buff[i+j*16]); 
					else sprintf(string, (const far rom char*)".");
					sio_puts(string);
				}
		}
		sprintf(string, (const far rom char*)"\n");
		sio_puts(string);
#endif
		//	-----------------------------------
		total_bytes = total_bytes + gpib_ptr;
		if(total_bytes >= 256) more = 0;
		if(rc==254) more = 0; // EOI
		//
		if(!WP_PORT) rc=pf_write((void*)gpib_buff, gpib_ptr, &nbytes);
		if(rc)
		{
			sprintf(string,(const far rom char*)"\r\nWrite fail %d", rc);
			sio_puts(string);
		}
		//sprintf(string,(const far rom char*)"\r\nWr returns %X, %d bytes wr", rc, nbytes);
		//sio_puts(string);
	}
	if(sector % 2 == 0)	// even
	{
		// Write the upper part of the 512-byte sector
		if(!WP_PORT) rc=pf_write((void*)buffer, 256, &nbytes);
	}
	if(!WP_PORT) rc = pf_write(0, 0, &nbytes);
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nClosing wr fail %d", rc);
		sio_puts(string);
	}
	else
	{
		//sprintf(string,(const far rom char*)"\r\nClosed sector");
		//sio_puts(string);
	}
#if SDEBUG > 0
	sprintf(string,(const far rom char*)"\r\nAM:Buff wr %ld bytes rx", total_bytes);
	sio_puts(string);
#endif
	amigo_increment();			// The address is incremented on a successful read.
	dsj = 0;
	EnablePP();
}
//
// DSJ, return status byte
//
void amigo_cmd_dsj()
{
#if SDEBUG > 0
	sprintf(string,(const far rom char*)"\r\nAM:DSJ");
	sio_puts(string);
#endif
	memset(gpib_tx_buff,0,GPIB_MAX_TX_LEN);
	//gpib_tx_buff[0] = ' ';
	//gpib_tx_buff[1] = 3;	// 2+1 byte
	//gpib_tx_buff[2] = 0;	// Status OK
	gpib_tx_buff[0] = dsj;
	sendgpib(1);	
	//
	// DSJ clears the DSJ byte that might be 2 after powerup
	//
	dsj = 0;
}

//
// Wakeup (undocumented), just enable PP response
//
//  54 3F 54 20 68 16 00 
//
void amigo_cmd_wakeup()
{
#if SDEBUG > 0
	sprintf(string,(const far rom char*)"\r\nAM:Wakeup");
	sio_puts(string);
#endif
	memset(gpib_tx_buff,0,GPIB_MAX_TX_LEN);
	//gpib_tx_buff[0] = ' ';
	//gpib_tx_buff[1] = 3;	// 2+1 byte
	//gpib_tx_buff[2] = 0;	// Status OK
	gpib_tx_buff[0] = 0;dsj;
	//sendgpib(1);
	EnablePP();
	//
	// Wakeup clears the DSJ byte that might be 2 after powerup
	//
	dsj = 0;
}
//
// Response to selected device clear or all clear - Select sector 0
//
//
void amigo_cmd_clear()
{
#if SDEBUG > 0
	sprintf(string,(const far rom char*)"\r\nAM:Clear");
	sio_puts(string);
#endif
	Disk.sector = 0;
	Disk.head = 0;
	Disk.cyl = 0;
	EnablePP();
	//
	// Clear the DSJ byte that might be 2 after powerup
	//
	dsj = 0;
}
//
// Increment sector (and possibly head and cylinder) then seek to that position
//
void amigo_increment()
{
DWORD fpos;
FRESULT rc;

	++Disk.sector;
	if(Disk.sector > 15)
	{
		Disk.sector = 0;
		++Disk.head;
		if(Disk.head > 1)
		{
			Disk.head = 0;
			++Disk.cyl;
		}
	}
	fpos = (DWORD)Disk.cyl*16L*256L*2L + (DWORD)Disk.head * 256L *16L + (DWORD)Disk.sector *256L;
#if SDEBUG > 0
	sprintf(string,(const far rom char*)"\r\nAM: Increment: %d %d %d %d, ", Disk.unit, Disk.cyl, Disk.head, Disk.sector);
	sio_puts(string);
	sprintf(string,(const far rom char*)"\r\nSeeking to %ld (%lX)", fpos, fpos);
	sio_puts(string);
#endif
	rc=pf_lseek(fpos);
	if(rc)
	{
		sprintf(string,(const far rom char*)"\r\nSeek to %ld failed %d", fpos, rc);
		sio_puts(string);
	}
	dsj = 0;
}