// Based on code from Raphael Kuhnen <raphael.kuhnen@fmf.uni-freiburg.de> and others.
//
// Credits: Ruud Baltissen - http://ftp.osuosl.org/pub/nslu2/sources/xmame-0.100.tar.bz2
// Credits: Ansgar Kueckes at hp9845.net for the finer points of the Amigo protocol
//
// PIC Pin addignments for GPIB
//
// RA0 - Parallell Poll status bit
// RA1 -
// RA2 -
// RA3 -
// RA4 -
// RA5 -
// RA6 - XTAL
// RA7 - XTAL
//
// RB0 - DAV
// RB1 - NFRD
// RB2 - NDAC
// RB3 - REN
// RB4 - ATN
// RB5 - EOI
// RB6 - IFC
// RB7 - SRQ
//
// RC0 -
// RC1 -
// RC2 -
// RC3 -
// RC4 -
// RC5 -
// RC6 - TX
// RC7 - RX
//
// RD0 - D1
// RD1 - D2
// RD2 - D3
// RD3 - D4
// RD4 - D5
// RD5 - D6
// RD6 - D7
// RD7 - D8
//
//
// GPIB Connector pinout
//
//Pin  	Name 	Signal Description 			Pin Name 	Signal Description 
//1 	DIO1 	Data Input/Output Bit 1 	13 	DIO5 	Data Input/Output Bit 5 
//2 	DIO2 	Data Input/Output Bit 2 	14 	DIO6 	Data Input/Output Bit 6 
//3 	DIO3 	Data Input/Output Bit 3 	15 	DIO7 	Data Input/Output Bit 7 
//4 	DIO4 	Data Input/Output Bit 4 	16 	DIO8 	Data Input/Output Bit 8 
//5 	EIO 	End-Or-Identify 			17 	REN 	Remote Enable 
//6 	DAV 	Data Valid 					18 	Shield 	Ground (DAV) 
//7 	NRFD 	Not Ready For Data 			19 	Shield 	Ground (NRFD) 
//8 	NDAC 	Not Data Accepted 			20 	Shield 	Ground (NDAC) 
//9 	IFC 	Interface Clear 			21 	Shield 	Ground (IFC) 
//10 	SRQ 	Service Request 			22 	Shield 	Ground (SRQ) 
//11 	ATN 	Attention 					23 	Shield 	Ground (ATN) 
//12 	Shield 	Chassis Ground 				24 	Single GND 	Single Ground

#include <p18f4620.h>
#include <stdio.h>
#include <string.h>
#include "defines.h"
#include "gpib.h"
#include "db_utils.h"
#include "sio.h"
#include "amigo.h"
#include "ss80.h"
#include "globals.h"

volatile char gpib_flags;
unsigned char eoiflag;
unsigned char atnflag;
unsigned char ifcflag;
unsigned char gpib_buff[GPIB_MAX_RX_LEN];
char gpib_tx_buff[GPIB_MAX_TX_LEN];
char string[MAX_DBG_STRING];
unsigned char edge[5];
unsigned int gpib_ptr;
//
#pragma udata access acs_scn
unsigned near char gpib_tx_ptr;
unsigned near int i,j;
unsigned near int count, timeout;

unsigned near char rx_state;				// Rx state machine state
unsigned near char tx_state;				// Tx state machine state

#pragma udata

unsigned char PP_BYTE=0xFF;

unsigned char idbyte1;
unsigned char idbyte2;

#pragma udata my_section_3
char gpib_tx_buff[GPIB_MAX_TX_LEN];
#pragma udata my_section_4
unsigned char gpib_buff[GPIB_MAX_RX_LEN];
#pragma udata

void DumpData(unsigned char *Ptr,int length);

void gpib_init()
{
	
	// REN high.
	REN_PORT = 1;
	DAV_LAT = 0;
	NRFD_LAT = 0;
	NDAC_LAT = 0;
	DAV_TRIS = 1;
	NRFD_TRIS = 1;
	NDAC_TRIS = 1;
	DIO_TRIS = 0xff;
	IFC_TRIS = 1;
	SRQ_TRIS = 1;
	ATN_TRIS = 1;
	EOI_TRIS = 1;
	REN_TRIS = 1;
	PP_TRIS = 0;		// PP helper circuitry, port = output

	//
 	// Zero the GPIB buffers and pointers
	//
	gpib_ptr = 0;
	gpib_tx_ptr = 0;
	i=0;
	while(i<GPIB_MAX_RX_LEN-1)
	{
		gpib_buff[i] = 0;
		i++;
	}

	i=0;
	while(i<GPIB_MAX_TX_LEN-1)
	{
		gpib_tx_buff[i] = 0;
		i++;
	}
	
	ifcflag = 0;
	return;
}

//-----------------------------------------------------------------------------------------------------------------------
// Send and receive routines are written as finite-state-machines. Please refer to the timing diagram on page 14 of
// "A tutorial Description of the Hewlett-Packard Interface Bus" (http://bitsavers.org/pdf/hp/hpib/TutorialDescrOfHPIB.pdf)
//-----------------------------------------------------------------------------------------------------------------------
//
// sendgpib(length, atn) - Send GPIB string in gpib_tx_buff[]
//
//
// If atn==1 then it is a command and ATN is set true (low)
// len is number of bytes to send.
// EOI is asserted on last character if EOIFLAG is set.
//
void sendgpib(int len, int atn)
{
	tx_state = GPIB_TX_CHECK;
	j=0;
	LATB = (LATB & 0b0001000);				// Do not touch REN. Not portable!
	//TRISB = (TRISB | 0b11110111);			// Do not touch REN.
	REN_TRIS = 1;

	DAV_LAT = 0;
	NRFD_LAT = 0;
	NDAC_LAT = 0;
	ATN_LAT = 0;
	DAV_TRIS = 1;
	NRFD_TRIS = 1;
	NDAC_TRIS = 1;
	ATN_TRIS = 1;
	DIO_TRIS = 0xff;
	//
	if(atn == 1) 				// Is it a command?
	{
		ATN_TRIS = 0;
		sprintf(string,(const far rom char*)"\r\nTXCMD len: %d ",len); // 
	} else
	{
		sprintf(string,(const far rom char*)"%S:%d ","\r\nTXDATA len", len);
	}
//#if defined(SDEBUG)
	sio_puts(string);
	//DumpData(&gpib_tx_buff[2],gpib_tx_buff[1]-2);
//#endif
	//
	TimeGPIB();
	while(tx_state != GPIB_TX_DONE && IFC_PORT == 1)
	{
		//sprintf(string,(const far rom char*)"\r\nTX State:%d  %X %X       ", tx_state ,PORTA, PORTB);
		//sio_puts(string);

#if defined(SDEBUG)
		if(ATN_PORT == 0 && EOI_PORT == 0)
		{
			sprintf(string,(const far rom char*)"{PP%d}",PP_PORT);	// 
			sio_puts(string);
		}
#endif

		//TimeGPIB(tx_state);
		switch(tx_state)
		{
//			case GPIB_TX_START:
//				tx_state = GPIB_TX_CHECK;
//				break;
			case GPIB_TX_CHECK:
				//if(PORTCbits.RC2 == 1 && PORTCbits.RC1 == 1)		// If both NRFD and NDAC high, then we have an error
				//	tx_state = GPIB_TX_ERROR;						// But we controllers do not care...
				//else
				//	tx_state = GPIB_TX_PUT_DATA;
				//break;
			case GPIB_TX_PUT_DATA:
#if defined(SDEBUG)
				sprintf(string,(const far rom char*)"%02X ", gpib_tx_buff[j] & 0xff); // Seems to need a small delay here...
																						// Removing this shaves 1300 ticks
				//sio_puts(string);
#endif
				if (j == len - 1 && (gpib_flags & FLAGS_EOI)!=0) // Last char and request eoi??
				{
					EOI_TRIS = 0;
//#if defined(SDEBUG)
					sprintf(string,(const far rom char*)"%S", "{EOI}");
					sio_puts(string);
//#endif
				}
				DIO_LAT = ~gpib_tx_buff[j];					// Put data on bus
				++j;

				DIO_TRIS = 0; 										// Make data lines output
																	//		A GPIB Talker, transfers data bytes to GPIB Listeners, it drives
																	//		the data byte on the GPIB DIO[81] signals. After waiting for a certain delay (known as
																	//		the T1 delay), the Talker asserts DAV to indicate to the Listeners that the data byte
																	//		has settled on the DIO[81] signals. A delay of 2us should satisfy most implementations.
																	//		http://www.ni.com/pdf/manuals/370875a.pdf
				//count = 2;
				//while(count){count--;} 								// wait to settle line
				tx_state = GPIB_TX_WAIT_FOR_NRFD;
				count = TIMEOUT;
				break;
			case GPIB_TX_WAIT_FOR_NRFD:								// Wait for NRFD high
				if (NRFD_PORT == 1 || count--<=0) tx_state = GPIB_TX_SET_DAV_LOW;

				break;
			case GPIB_TX_SET_DAV_LOW:
				//		DAV = 0;
				DAV_TRIS = 0;
				tx_state = GPIB_TX_WAIT_FOR_NDAC;
				count = TIMEOUT;
				break;
			case GPIB_TX_WAIT_FOR_NDAC:
				if(NDAC_PORT == 1 || count--<=0) tx_state = GPIB_TX_SET_DAV_HIGH;
				break;
			case GPIB_TX_SET_DAV_HIGH:
				DIO_TRIS=0xff; 										// data lines tristate
				EOI_TRIS = 1; 										// EOI false
				//		DAV = 1;
				DAV_TRIS = 1;										// DAV high
				//
				// More data?
				//
				if (j < len) tx_state = GPIB_TX_CHECK;
				else tx_state = GPIB_TX_DONE;
				break;
			case GPIB_TX_ERROR:
				sprintf(string,(const far rom char*)"%S", "{ERR}");
				sio_puts(string);
				tx_state = GPIB_TX_DONE;
				break;
		}
	}
	// All sent
	if(atn == 1) {ATN_TRIS = 1;} // ATN = 1								
	DAV_TRIS = 1;													// DAV, NRFD etc threestate (input)
	NRFD_TRIS = 1;
	NDAC_TRIS = 1;
	ATN_TRIS = 1;
	EOI_TRIS = 1;												
	DIO_TRIS = 0xff;												// Data threestate (input)
	TRISB = (TRISB | 0b11110111);									// Do not touch REN. Not portable!
	TimeGPIB();
}

//
// Receive a string of maximum GPIB_MAX_LEN chars from GPIB. Routine will loop forever and only exit when it has seen an EOI 
// or if the buffer is full. 
//
// returns the actual number of bytes received in gpib_ptr.
// Returncode is an Amigo or SS/80 command or 255 if timeout occurred.
//
int readgpib(void)
{
unsigned char timeoutflag=0;
int status;

	timeout = 2000;
	gpib_ptr=0;
	status = 0;
	//LATB = (LATB & 0b00001000);			// Do not touch REN.	Not portable
	LATB = 0;
	//TRISB = (TRISB | 0b11110111);			// Do not touch REN.
	REN_TRIS = 1;
	DAV_LAT = 0;
	NRFD_LAT = 0;
	NDAC_LAT = 0;
	NRFD_PORT=0;
	NDAC_PORT=0;
	DAV_TRIS = 1;
	NRFD_TRIS = 0;
	NDAC_TRIS = 0;
	DIO_TRIS = 0xff;

	rx_state = GPIB_RX_START;
	memset(gpib_buff,0,GPIB_MAX_RX_LEN); // Init buffer to a known state

	while(rx_state != GPIB_RX_DONE && IFC_PORT == 1)
	{
		//sprintf(string,(const far rom char*)"RX State:%d %d %X %X       ", rx_state,gpib_ptr ,PORTD, PORTB);
		//sio_puts((unsigned char*)string);
		//
		if(gpib_ptr>=GPIB_MAX_RX_LEN)
		{										// Break out if Buffer full
			rx_state = GPIB_RX_FINISH;
		}
		//
		// Parallel poll? 
		//
		// According to the SS 80 manual, the device with ID0 should pull data line DIO8 low to indicate completion
		//
		// Th disk drive manual says for the DSJ command for instance:
		// Parallel Poll is disabled after the reception of the the secondary and
		// is not re-enabled after the completion of the command.
		//
		// The circuitry has two alternative ways to handle this. One is software only, if SOFTWARE_PP is enabled. The other is
		// hardware, where ATN and EOI are fed into an NOR gate, together with an MPU pin. PP happens when both ATN and EOI are
		// active (low) and when the MPU pin is low, the output of the NOR gate goes high, driving a mosfet that pulls DIO8 low, to
		// indicate a positive PP response from unit 0
		//
		if(ATN_PORT == 0 && EOI_PORT == 0)
		{
#if defined(SOFTWARE_PP)
			DIO_LAT = 0x00;								// Put status bit on bus
			DIO_TRIS=PP_BYTE; 								// Just bit 7 (DIO8) active for now
			while (ATN_PORT == 0 && EOI_PORT == 0)
					;									// Wait for PP to go away
#if defined(SDEBUG)
			sprintf(string,(const far rom char*)"<PP%d>",PP_BYTE);	// Need a small delay..
			sio_puts((unsigned char*)string);
#endif
			DIO_TRIS=0xff; 
#else
			sprintf(string,(const far rom char*)"<PP%d>",PP_PORT);	// Need a small delay..
			sio_puts(string);
#endif
		}	
		switch(rx_state)
		{
			case GPIB_RX_START:
				//		NRFD = 0;
				NRFD_TRIS = 0;							// Set both NRFD and NDAC low
				NDAC_TRIS = 0;
				rx_state = GPIB_RX_ACCEPT;
				break;
	
			case GPIB_RX_ACCEPT:

				//		NRFD = 1;		
				NRFD_TRIS = 1;							// Set NRFD high
				timeout = 2000;
				rx_state = GPIB_RX_WAIT_DAV;
				//
				break;
	
			case GPIB_RX_WAIT_DAV:						// Wait here for DAV to go low
				if (DAV_PORT == 0) rx_state = GPIB_RX_DAV_LOW;
				--timeout;
				if(timeout <=0)	 rx_state = GPIB_RX_DAV_TIMEOUT;

				break;
	
			case GPIB_RX_DAV_LOW:
				//		NRFD = 1;
				timeout = 2000;		
				NRFD_TRIS = 0;							// Set NRFD low
				gpib_buff[gpib_ptr] = ~DIO_PORT;		// Accept databyte
				//
				eoiflag = EOI_PORT;
				atnflag = ATN_PORT;
				//

#if defined(SDEBUG1)
				sprintf(string,(const far rom char*)"<%02X>", gpib_buff[gpib_ptr]);
				sio_puts((unsigned char*)string);
#endif
				gpib_ptr++;
				//
				count = 10;
	
				NDAC_TRIS = 1;							// Set NDAC high
				rx_state = GPIB_RX_WAIT_DAV_HIGH;
				break;
	
			case GPIB_RX_WAIT_DAV_HIGH:					// Wait here for DAV to go high
				if (DAV_PORT == 1) rx_state = GPIB_RX_DAV_HIGH;
				break;
	
			case GPIB_RX_DAV_HIGH:
				NDAC_TRIS = 0;							// Set NDAC low
				//
				// Analyze, Call appropriate handler for Amigor or SS/80
				// status <> 0 means we have an amigo or SS80-command waiting
				//
				if(protocol)
				{
					status = SS80_char(gpib_buff[gpib_ptr-1],!atnflag,ifcflag);
				}
				else
				{
				status = amigo_char(gpib_buff[gpib_ptr-1],!atnflag,ifcflag);
				}

				if(status) rx_state = GPIB_RX_FINISH;	// Command waiting?
				else if (eoiflag==0)					// seen an EOI? EOI is set on the last byte of data message
				{
					rx_state = GPIB_RX_EOI;
					count = 1000;						// Prime timeout counter
#if defined(SDEBUG)
					sprintf(string,(const far rom char*)"<EOI>");
					sio_puts(string);
#endif
					status = 254; //EOIFLAG2;
					break;
				}
				else rx_state = GPIB_RX_ACCEPT;						// no,  back to the top
				break;
	
			case GPIB_RX_EOI:							// Seen an EOI, ack and wait here until it goes away or timeout
	
				NDAC_TRIS = 1;							// Set NDAC high
				if (EOI_PORT == 1) rx_state = GPIB_RX_FINISH;
				count--;
				if (count <=0) rx_state = GPIB_RX_FINISH;
				break;


			case GPIB_RX_DAV_TIMEOUT:
				//sprintf(string,(const far rom char*)"[TO]");
				//sio_puts((unsigned char*)string);
				timeoutflag=1;
				rx_state = GPIB_RX_FINISH;
				break;
			case GPIB_RX_FINISH:
				NRFD_TRIS = 0;							// Set both NRFD and NDAC low
				NDAC_TRIS = 0;
				rx_state = GPIB_RX_DONE;
				break;
		}

	} // while
	//
	if(atnflag == 0 && gpib_buff[gpib_ptr-1] == 0x60 && gpib_buff[gpib_ptr-2] == 0x5f)	// GPIB Ident sequence?
	{
		sprintf(string,(const far rom char*)"<ident>");
		sio_puts(string);
		gpib_flags = gpib_flags | FLAGS_EOI;
		gpib_tx_buff[0] = idbyte1;			// Identify ourselves as a HP 9121 drive
		gpib_tx_buff[1] = idbyte2;			// $0104 or 9122 $0222
		sendgpib(2,0);
		gpib_flags = gpib_flags & ~FLAGS_EOI;
		gpib_ptr = 0;
	}
	if(gpib_ptr > 0)
	{
		//gpib_buff[gpib_ptr] = 0;
		//sprintf(string,(const far rom char*)"\nRX: %d",gpib_ptr);
		//sio_puts((unsigned char*)string);
	}
	if(timeoutflag)
	{
		return 255;
	}
	else
	{
		return status;
	}
}

//
// Interface clear
//
void gpib_ifc(void)
{
	IFC_PORT = 0;
	IFC_TRIS = 0;
	//
	// Hold IFC low for >100us
	count = TIMEOUT;
	while (count)
	{
		count--;
	}
	IFC_PORT = 1;
	IFC_TRIS = 1;
	count = TIMEOUT;
	while (count)
	{
		count--;
	}
}

//
// Assert REN to put instrument in remote mode
// state = true -> assert REN, state = false -> deasert REN
//
void gpib_ren(unsigned char state)
{
	if(state == true)
	{
		REN_TRIS = 0;
		REN_PORT = 0;
	}
	else
	{
		REN_TRIS = 1;
		REN_PORT = 1;
	}
}


void TimeGPIB()
{
unsigned int ticks = TMR0L + 256 * TMR0H;
	sprintf(string,(const far rom char*)"\r\nT:%u", ticks);
	sio_puts(string);
	INTCONbits.TMR0IF = 0;
}